import userEvent from "@testing-library/user-event";

import { createMockMetadata } from "__support__/metadata";
import { renderWithProviders, screen } from "__support__/ui";
import MetabaseSettings from "metabase/lib/settings";
import { checkNotNull } from "metabase/lib/types";
import type { ClickActionProps } from "metabase/visualizations/types";
import type { DatasetColumn } from "metabase-types/api";
import {
  createSampleDatabase,
  PEOPLE,
  PEOPLE_ID,
} from "metabase-types/api/mocks/presets";

import { AutomaticInsightsDrill } from "./AutomaticInsightsDrill";

describe("AutomaticInsightsDrill", () => {
  describe('"enable-xrays" feature is disabled', () => {
    beforeAll(() => {
      MetabaseSettings.set("enable-xrays", false);
    });

    afterAll(() => {
      MetabaseSettings.set("enable-xrays", true);
    });

    it("should return empty array", () => {
      const actions = AutomaticInsightsDrill(setupCategoryFieldQuery());

      expect(actions).toHaveLength(0);
    });
  });

  it("should return one item with popover", () => {
    const actions = AutomaticInsightsDrill(setupCategoryFieldQuery());

    expect(actions).toHaveLength(1);

    const { title, popover } = actions[0];
    expect(title).toBe("Automatic insights…");
    expect(popover).toBeTruthy();
  });

  describe("drill popover", () => {
    it("should render title and options", () => {
      setup();

      expect(screen.getByText("Automatic insights…")).toBeInTheDocument();
      expect(screen.getByText("Compare to the rest")).toBeInTheDocument();
      expect(screen.getByText("X-ray")).toBeInTheDocument();
    });

    it(`should return "url" prop with a relative link on action item click`, () => {
      const stringStartingWithWordChar = /^\w+/;
      const {
        props: { onClick },
      } = setup();

      userEvent.click(screen.getByText("Compare to the rest"));
      expect(onClick).toHaveBeenCalledTimes(1);

      expect(onClick.mock.lastCall[0].url()).toMatch(
        stringStartingWithWordChar,
      );

      userEvent.click(screen.getByText("X-ray"));
      expect(onClick).toHaveBeenCalledTimes(2);

      expect(onClick.mock.lastCall[0].url()).toMatch(
        stringStartingWithWordChar,
      );
    });
  });
});

function setupCategoryFieldQuery(): ClickActionProps {
  const metadata = createMockMetadata({
    databases: [createSampleDatabase()],
  });

  const peopleTable = checkNotNull(metadata.table(PEOPLE_ID));

  const query = peopleTable
    .query()
    .aggregate(["count"])
    .breakout(["field", PEOPLE.STATE, null]);

  const question = query.question();

  const clicked = {
    column: query.aggregationDimensions()[0].column() as DatasetColumn,
    value: 194,
    dimensions: [
      {
        column: checkNotNull(metadata.field(PEOPLE.STATE))
          .dimension()
          .column() as DatasetColumn,
        value: "TX",
      },
    ],
  };

  return { question, clicked };
}

function setup() {
  const actions = AutomaticInsightsDrill(setupCategoryFieldQuery());
  expect(actions).toHaveLength(1);

  const { popover: PopoverComponent } = actions[0];

  const props = {
    series: [],
    onClick: jest.fn(),
    onChangeCardAndRun: jest.fn(),
    onUpdateVisualizationSettings: jest.fn(),
    onResize: jest.fn(),
    onClose: jest.fn(),
  };

  renderWithProviders(<PopoverComponent {...props} />);

  return { props };
}
